/*
 * ********************************************************************************************************************
 *
 *                      :*$@@%$*:                         ;:                ;;    ;;
 *                    :@@%!  :!@@%:                       %!             ;%%@@%$ =@@@@@@@%;     @%@@@%%%%@@@@@
 *                   :@%;       :$=                       %%$$$%$$         ;$$  ;$@=   !@$
 *                   =@!                                  %!              @ $=;%   !@@@%:      !$$$$$$$$$$$$$$=
 *                   =@*                                  %!              @ $= % %@=   =%@!      %=
 *              *$%%! @@=        ;=$%%%$*:                %!              @ $= % =%%%%%%@$      *%:         =%
 *            %@@!:    !@@@%=$@@@@%!  :*@@$:              %!              @ $= % $*     ;@      @*          :%*
 *          ;@@!          ;!!!;:         ;@%:      =======@%========*     @ $$ % $%*****$@     :@$=*********=@$
 *          $@*   ;@@@%=!:                *@*
 *          =@$    ;;;!=%@@@@=!           =@!
 *           %@$:      =@%: :*@@@*       %@=                    Copyright (c) 2016-2020.  北京上格云技术有限公司
 *            ;%@@$=$@@%*       *@@@$=%@@%;
 *               ::;::             ::;::                                              All rights reserved.
 *
 * ********************************************************************************************************************
 */

import { SGraphItem, SGraphScene } from "@saga-web/graph/lib";
import { Equip } from "./types/Equip";
import { Relation } from "./types/Relation";
import { EquipItem } from "./items/EquipItem";
import { RelationItem } from "./items/RelationItem";
import { SMouseEvent } from "@saga-web/base/lib";
import { Anchor } from "./types/Anchor";
import { SMathUtil } from "./util/SMathUtil";
import { AnchorItem } from "./items/AnchorItem";
import { GraphyData } from "./types/GraphyData";
import { DirectRelationItem } from "./items/DirectRelationItem";
import { BaseRelationItem } from "./items/BaseRelationItem";
import { LineType } from "./types/LineType";

/**
 * 拓扑图场景
 *
 * @author  郝建龙
 */
export class TopoScene extends SGraphScene {
    /** 所有设备    */
    EquipList: EquipItem[] = [];
    /** 所有关系item    */
    RelationList: BaseRelationItem[] = [];
    /** 所有锚点    */
    AnchorList: AnchorItem[] = [];
    /** 设备是否隐藏  */
    _isShowEquip: boolean = true;
    get isShowEquip(): boolean {
        return this._isShowEquip;
    } // Get isShowEquip
    set isShowEquip(v: boolean) {
        if (this._isShowEquip === v) {
            return;
        }
        this._isShowEquip = v;
        this.EquipList.map(
            (t: EquipItem): EquipItem => {
                t.visible = this._isShowEquip;
                return t;
            }
        );
    } // Set isShowEquip

    /** 设备是否隐藏  */
    _isShowRelation: boolean = true;
    get isShowRelation(): boolean {
        return this._isShowRelation;
    } // Get isShowEquip
    set isShowRelation(v: boolean) {
        if (this._isShowRelation === v) {
            return;
        }
        this._isShowRelation = v;
        this.RelationList.map(
            (t: BaseRelationItem): BaseRelationItem => {
                t.visible = this._isShowRelation;
                return t;
            }
        );
    } // Set isShowEquip

    /** 添加item状态    */
    isAdding: boolean = false;
    /** 划线类型    */
    lineType: number = LineType.None;
    /** 当前关系item    */
    relatItem: RelationItem | DirectRelationItem | null = null;
    /** 设备-关系item统一回调   */
    callback: Function = (): void => {};
    /** */
    loadData(data: GraphyData): void {
        if (data.equipList && data.equipList.length) {
            this.addAllEquip(data.equipList);
        }
        if (data.relationList && data.relationList.length) {
            this.addAllRelation(data.relationList);
        }
    } // Function loadData()

    /**
     * 添加所有设备item
     *
     */
    addAllEquip(list: Equip[]): void {
        if (list.length) {
            list.forEach(t => {
                this.addEquip(t);
            });
        }
    } // Function addAllEquip()

    /**
     *  添加设备item
     */
    addEquip(equip: Equip): void {
        let item = new EquipItem(null, equip);
        item.moveTo(equip.Pos.X, equip.Pos.Y);
        this.click(item);
        item.visible = this.isShowEquip;
        this.EquipList.push(item);
        if (equip.AnchorList && equip.AnchorList.length) {
            this.addAllAnchor(item, equip.AnchorList);
        }
        this.addItem(item);
    } // Function addEquip()

    /**
     * 添加单个设备item的锚点
     *
     */
    addAllAnchor(parent: SGraphItem, list: Anchor[]): void {
        if (list.length) {
            list.forEach(t => {
                this.addAnchor(parent, t);
            });
        }
    } // Function addAllAnchor()

    /**
     *  添加设备item的锚点
     */
    addAnchor(parent: SGraphItem, anchor: Anchor): void {
        let item = new AnchorItem(parent, anchor);
        this.click(item);
        item.visible = this.isShowEquip;
        this.AnchorList.push(item);
    } // Function addAnchor()

    /**
     * 添加所有设备item
     *
     */
    addAllRelation(list: Relation[]): void {
        if (list.length) {
            list.forEach(t => {
                this.addRelation(t);
            });
        }
    } // Function addAllRelation()

    /**
     *  添加设备item
     */
    addRelation(rela: Relation): void {
        let item = new RelationItem(null, rela);
        item.visible = this.isShowRelation;
        this.click(item);
        this.RelationList.push(item);
        this.addItem(item);
    } // Function addRelation()

    /**
     *  点击按下事件
     *
     */
    onMouseDown(event: SMouseEvent): boolean {
        if (this.isAdding) {
            return this.addSelectItem(event);
        } else if (this.lineType > 0) {
            let anc = this.clickIsAnchor(event);
            if (anc) {
                // 点击的是锚点,直接处理添加到关系item点数据中
                let ancPoint = anc.mapToScene(anc.X, anc.Y);
                // 将点击到的锚点状态置为已连接
                anc.isConnected = true;
                if (this.relatItem) {
                    console.log("第二次点击锚点");
                    this.relatItem.Anchor2 = anc;
                    if (anc.parent) {
                        anc.parent.connect(
                            "changePos",
                            this.relatItem,
                            this.relatItem.change
                        );
                    }
                    event.x = ancPoint.x;
                    event.y = ancPoint.y;
                    this.lineType = LineType.None;
                    this.relatItem.onMouseDown(event);
                    this.relatItem = null;
                    this.grabItem = null;
                    return false;
                } else {
                    let data = {
                        Anchor1: anc
                    };
                    let item;
                    if (this.lineType == LineType.Polygonal) {
                        //绘制折线
                        item = new RelationItem(null, data);
                    } else if (this.lineType == LineType.Curve) {
                        //绘制曲线--TODO
                        item = new RelationItem(null, data);
                    } else if (this.lineType == LineType.Straight) {
                        //绘制直线
                        item = new DirectRelationItem(null, data);
                    }
                    if (item) {
                        this.addItem(item);
                        if (anc.parent) {
                            anc.parent.connect("changePos", item, item.change);
                        }
                        this.click(item);
                        this.RelationList.push(item);
                        this.relatItem = item;
                        this.grabItem = item;
                    }
                    return true;
                }
            } else {
                console.log("未点击锚点");
                // 点击的不是锚点,处理点击点与
                if (this.relatItem) {
                    return this.relatItem.onMouseDown(event);
                }
                return false;
            }
        } else {
            return super.onMouseDown(event);
        }
    } // Function onClick

    /**
     *  点击移动事件
     *
     */
    onMouseMove(event: SMouseEvent): boolean {
        if (this.lineType && this.relatItem) {
            this.relatItem.onMouseMove(event);
        } else {
            super.onMouseMove(event);
        }
        return false;
    }

    /**
     * 根据点击设备创建对应item
     *
     * */
    createItem(data: Equip): SGraphItem {
        this.isAdding = true;
        this.grabItem = new EquipItem(null, data);
        this.click(this.grabItem);
        return this.grabItem;
    } // Function createItem()

    /**
     * 根据当前选中的item添加到场景
     *
     * @param event 点击事件
     * @return 是否添加成功
     * */
    addSelectItem(event: SMouseEvent): boolean {
        if (this.grabItem && this.grabItem instanceof EquipItem) {
            // @ts-ignore
            this.grabItem.data.Pos = {};
            // @ts-ignore
            this.grabItem.data.Pos.X = event.x;
            // @ts-ignore
            this.grabItem.data.Pos.Y = event.y;
            this.grabItem.moveTo(event.x, event.y);
            this.addItem(this.grabItem);
            this.EquipList.push(this.grabItem);
            if (this.grabItem.data && this.grabItem.data.AnchorList.length) {
                this.addAllAnchor(this.grabItem, this.grabItem.data.AnchorList);
            }
            this.isAdding = false;
            this.grabItem = null;
            return true;
        }
        return false;
    } // Function addSelectItem()

    /**
     *  划线时点击位置是否是锚点
     *
     *  @param  event   事件
     *  @param  len     限制距离
     *  @return 点击的锚点
     * */
    clickIsAnchor(event: SMouseEvent, len: number = 25): AnchorItem | null {
        let minAnchor = null;
        if (this.view) {
            len = len / this.view.scale;
        }
        this.AnchorList.forEach(anchor => {
            if (anchor.data) {
                let anchorPoint = anchor.mapToScene(anchor.X, anchor.Y);
                let dis = SMathUtil.pointDistance(
                    event.x,
                    event.y,
                    anchorPoint.x,
                    anchorPoint.y
                );
                if (dis < len) {
                    minAnchor = anchor;
                    len = dis;
                }
            }
        });
        return minAnchor;
    } // Function clickIsAnchor()

    /**
     *  绑定设备右键事件
     *
     */
    click(item: SGraphItem): void {
        item.connect("ContextMenu", this, this.callback);
    }
} // Class TopoScene
